import json
import logging
import os
from abc import ABC, abstractmethod
from typing import List, Optional

from ezbeq.catalogue import Catalogue
from ezbeq.config import Config

logger = logging.getLogger('ezbeq.device')


class SlotState:

    def __init__(self, slot_id: str):
        self.__slot_id = slot_id
        self.last = 'Empty'
        self.active = False

    @property
    def slot_id(self) -> str:
        return self.__slot_id

    def merge_with(self, state: dict) -> bool:
        if 'last' in state:
            self.last = state['last']
        return True

    def as_dict(self) -> dict:
        return {'id': self.slot_id, 'last': self.last, 'active': self.active}

    def __repr__(self):
        return f"{'*' if self.active else ''} {self.slot_id} - {self.last}"

    def clear(self):
        self.last = 'Empty'

    def set_gain(self, channel: Optional[int], value: float):
        pass

    def mute(self, channel: Optional[int]):
        pass

    def unmute(self, channel: Optional[int]):
        pass


class DeviceControls:

    def __init__(self):
        self.mute: bool = False
        self.master_volume: float = 0.0

    def as_dict(self) -> dict:
        return {
            'mute': self.mute,
            'masterVolume': self.master_volume
        }

    def __repr__(self):
        return f"MV: {self.master_volume:.2f} Mute: {self.mute}"


class DeviceState:

    def __init__(self, name: str):
        self.__name: str = name
        self.__slots: List[SlotState] = []
        self.__device_controls: Optional[DeviceControls] = None

    @property
    def name(self):
        return self.__name

    def __repr__(self):
        return f"{self.name} {self.__device_controls if self.__device_controls else ''} [Slots: {self.__slots}]"

    @property
    def slots(self) -> List[SlotState]:
        return self.__slots

    @slots.setter
    def slots(self, slots: List[SlotState]):
        assert self.__slots == [], 'can only initialise slots once'
        assert slots != [], 'must have slots'
        self.__slots = slots
        # TODO fix
        from ezbeq.minidsp import MinidspSlotState
        if isinstance(slots[0], MinidspSlotState):
            self.__device_controls = DeviceControls()

    @property
    def master_volume(self) -> Optional[float]:
        return self.__device_controls.master_volume if self.__device_controls else None

    @master_volume.setter
    def master_volume(self, mv: float):
        if self.__device_controls:
            self.__device_controls.master_volume = mv

    @property
    def mute(self) -> Optional[bool]:
        return self.__device_controls.mute if self.__device_controls else None

    @mute.setter
    def mute(self, mute: bool):
        if self.__device_controls:
            self.__device_controls.mute = mute

    def merge_with(self, serialised_state: dict) -> bool:
        saved_slots_by_id = {v['id']: v for v in serialised_state}
        current_slots_by_id = {s.slot_id: s for s in self.__slots}
        if saved_slots_by_id.keys() == current_slots_by_id.keys():
            for slot_id, state in saved_slots_by_id.items():
                current_slots_by_id[slot_id].merge_with(state)
            return True
        else:
            return False

    def get_slot(self, slot_id) -> SlotState:
        return next(s for s in self.__slots if s.slot_id == slot_id)

    def serialise(self):
        return [s.as_dict() for s in self.__slots]

    def as_dict(self) -> dict:
        d_dict = self.__device_controls.as_dict() if self.__device_controls else {}
        return {
            **d_dict,
            'slots': [s.as_dict() for s in self.__slots]
        }

    def activate(self, slot_id: str):
        for s in self.__slots:
            s.active = s.slot_id == slot_id


class DeviceStateHolder:

    def __init__(self, cfg: Config):
        self.__state = DeviceState('master')
        self.__cached_state = {}
        self.__device_type = None
        self.__file_name = os.path.join(cfg.config_path, 'device.json')
        self.__initialised = False
        if os.path.exists(self.__file_name):
            with open(self.__file_name, 'r') as f:
                self.__cached_state = json.load(f)
            logger.info(f"Loaded {self.__cached_state} from {self.__file_name}")
        else:
            logger.info(f"No cached state found at {self.__file_name}")

    def initialise(self, bridge: 'DeviceBridge') -> None:
        if not self.__initialised:
            self.__device_type = bridge.device_type()
            self.__state.slots = bridge.slot_state()
            if self.__cached_state:
                if not self.__state.merge_with(self.__cached_state):
                    logger.warning(f"Discarded {self.__cached_state} from {self.__file_name}, does not match {self.__state}")
            device_state = bridge.state()
            if device_state:
                if 'active_slot' in device_state:
                    self.activate(device_state['active_slot'])
                if 'mute' in device_state:
                    self.master_mute = device_state['mute'] is True
                if 'volume' in device_state:
                    self.master_volume = device_state['volume']
            self.__initialised = True

    def activate(self, slot: str):
        self.__state.activate(slot)

    def set_loaded_entry(self, slot: str, entry: Catalogue):
        self.__set_last(slot, entry.formatted_title)

    def __set_last(self, slot: str, title: str):
        self.__state.get_slot(slot).last = title
        self.__activate_and_cache(slot)

    def __activate_and_cache(self, slot: str):
        self.activate(slot)
        with open(self.__file_name, 'w') as f:
            json.dump(self.__state.serialise(), f, sort_keys=True)

    def error(self, slot: str):
        self.__set_last(slot, 'ERROR')
        self.__activate_and_cache(slot)

    def clear(self, slot: str):
        self.__state.get_slot(slot).clear()
        self.__activate_and_cache(slot)

    @property
    def master_volume(self):
        return self.__state.master_volume

    @master_volume.setter
    def master_volume(self, value: float):
        self.__state.master_volume = value

    @property
    def master_mute(self):
        return self.__state.mute

    @master_mute.setter
    def master_mute(self, value: bool):
        self.__state.mute = value

    def set_slot_gain(self, slot: str, channel: Optional[int], value: float):
        self.__state.get_slot(slot).set_gain(channel, value)
        self.__activate_and_cache(slot)

    def mute_slot(self, slot: str, channel: Optional[int]):
        self.__state.get_slot(slot).mute(channel)
        self.__activate_and_cache(slot)

    def unmute_slot(self, slot: str, channel: Optional[int]):
        self.__state.get_slot(slot).unmute(channel)
        self.__activate_and_cache(slot)

    def get(self) -> DeviceState:
        return self.__state


class Bridge(ABC):

    @abstractmethod
    def device_type(self) -> str:
        pass

    @abstractmethod
    def slot_state(self) -> List[SlotState]:
        pass

    @abstractmethod
    def state(self) -> Optional[dict]:
        pass

    @abstractmethod
    def activate(self, slot: str) -> None:
        pass

    @abstractmethod
    def load_filter(self, slot: str, entry: Catalogue) -> None:
        pass

    @abstractmethod
    def clear_filter(self, slot: str) -> None:
        pass

    @abstractmethod
    def mute(self, slot: Optional[str], channel: Optional[int]) -> None:
        pass

    @abstractmethod
    def unmute(self, slot: Optional[str], channel: Optional[int]) -> None:
        pass

    @abstractmethod
    def set_gain(self, slot: Optional[str], channel: Optional[int], gain: float) -> None:
        pass


class DeviceBridge(Bridge):

    def __init__(self, cfg: Config):
        self.__bridge = create_bridge(cfg)

    def device_type(self):
        return self.__bridge.device_type()

    def slot_state(self) -> List[SlotState]:
        return self.__bridge.slot_state()

    def supports_gain(self):
        from ezbeq.minidsp import Minidsp
        return isinstance(self.__bridge, Minidsp)

    def state(self) -> Optional[dict]:
        return self.__bridge.state()

    def activate(self, slot: str):
        return self.__bridge.activate(slot)

    def load_filter(self, slot: str, entry: Catalogue):
        return self.__bridge.load_filter(slot, entry)

    def clear_filter(self, slot: str):
        return self.__bridge.clear_filter(slot)

    def mute(self, slot: Optional[str], channel: Optional[int]) -> None:
        return self.__bridge.mute(slot, channel)

    def unmute(self, slot: Optional[str], channel: Optional[int]) -> None:
        return self.__bridge.unmute(slot, channel)

    def set_gain(self, slot: Optional[str], channel: Optional[int], gain: float) -> None:
        return self.__bridge.set_gain(slot, channel, gain)


def create_bridge(cfg: Config) -> Bridge:
    if cfg.minidsp_exe:
        from ezbeq.minidsp import Minidsp
        return Minidsp(cfg)
    elif cfg.htp1_options:
        from ezbeq.htp1 import Htp1
        return Htp1(cfg)
    elif cfg.jriver_options:
        from ezbeq.jriver import JRiver
        return JRiver(cfg)
    else:
        raise ValueError('No device configured')


class InvalidRequestError(Exception):
    pass



import logging
import os
from concurrent.futures.thread import ThreadPoolExecutor
from contextlib import contextmanager
from typing import List, Optional, Union

import time

from ezbeq.catalogue import Catalogue
from ezbeq.config import Config
from ezbeq.device import Bridge, InvalidRequestError, SlotState

logger = logging.getLogger('ezbeq.minidsp')


class Minidsp(Bridge):

    def __init__(self, cfg: Config):
        self.__executor = ThreadPoolExecutor(max_workers=1)
        self.__cmd_timeout = cfg.minidsp_cmd_timeout
        self.__ignore_retcode = cfg.ignore_retcode
        self.__runner = cfg.create_minidsp_runner()

    def device_type(self) -> str:
        return 'minidsp'

    def slot_state(self) -> List[SlotState]:
        return [MinidspSlotState(c_id) for c_id in [str(i + 1) for i in range(4)]]

    def state(self) -> Optional[dict]:
        return self.__executor.submit(self.__get_state).result(timeout=self.__cmd_timeout)

    def __get_state(self) -> Optional[dict]:
        values = {}
        lines = None
        try:
            kwargs = {'retcode': None} if self.__ignore_retcode else {}
            lines = self.__runner(timeout=self.__cmd_timeout, **kwargs)
            for line in lines.split('\n'):
                if line.startswith('MasterStatus'):
                    idx = line.find('{ ')
                    if idx > -1:
                        vals = line[idx + 2:-2].split(', ')
                        for v in vals:
                            if v.startswith('preset: '):
                                values['active_slot'] = str(int(v[8:]) + 1)
                            elif v.startswith('mute: '):
                                values['mute'] = v[6:] == 'true'
                            elif v.startswith('volume: Gain('):
                                values['volume'] = float(v[13:-1])
        except:
            logger.exception(f"Unable to locate active preset in {lines}")
        return values

    @staticmethod
    def __as_idx(idx: Union[int, str]):
        return int(idx) - 1

    def __send_cmds(self, target_slot_idx: Optional[int], cmds: List[str]):
        return self.__executor.submit(self.__do_run, cmds, target_slot_idx).result(timeout=self.__cmd_timeout)

    def activate(self, slot: str) -> None:
        target_slot_idx = self.__as_idx(slot)
        self.__validate_slot_idx(target_slot_idx)
        self.__send_cmds(target_slot_idx, [])

    @staticmethod
    def __validate_slot_idx(target_slot_idx):
        if target_slot_idx < 0 or target_slot_idx > 3:
            raise InvalidRequestError(f"Slot must be in range 1-4")

    def load_filter(self, slot: str, entry: Catalogue) -> None:
        target_slot_idx = self.__as_idx(slot)
        self.__validate_slot_idx(target_slot_idx)
        cmds = MinidspBeqCommandGenerator.filt(entry)
        self.__send_cmds(target_slot_idx, cmds)

    def clear_filter(self, slot: str) -> None:
        target_slot_idx = self.__as_idx(slot)
        self.__validate_slot_idx(target_slot_idx)
        cmds = MinidspBeqCommandGenerator.filt(None)
        cmds.extend(MinidspBeqCommandGenerator.mute(False, target_slot_idx, None))
        cmds.extend(MinidspBeqCommandGenerator.gain(0.0, target_slot_idx, None))
        self.__send_cmds(target_slot_idx, cmds)

    def mute(self, slot: Optional[str], channel: Optional[int]) -> None:
        self.__do_mute_op(slot, channel, True)

    def __do_mute_op(self, slot: Optional[str], channel: Optional[int], state: bool):
        target_channel_idx, target_slot_idx = self.__as_idxes(channel, slot)
        if target_slot_idx:
            self.__validate_slot_idx(target_slot_idx)
        cmds = MinidspBeqCommandGenerator.mute(state, target_slot_idx, target_channel_idx)
        self.__send_cmds(target_slot_idx, cmds)

    def unmute(self, slot: Optional[str], channel: Optional[int]) -> None:
        self.__do_mute_op(slot, channel, False)

    def set_gain(self, slot: Optional[str], channel: Optional[int], gain: float) -> None:
        target_channel_idx, target_slot_idx = self.__as_idxes(channel, slot)
        cmds = MinidspBeqCommandGenerator.gain(gain, target_slot_idx, target_channel_idx)
        self.__send_cmds(target_slot_idx, cmds)

    def __as_idxes(self, channel, slot):
        target_slot_idx = self.__as_idx(slot) if slot else None
        target_channel_idx = self.__as_idx(channel) if channel else None
        return target_channel_idx, target_slot_idx

    def __do_run(self, config_cmds: List[str], slot: Optional[int]):
        if slot is not None:
            change_slot = True
            current_state = self.__get_state()
            if current_state and 'active_slot' in current_state and int(current_state['active_slot']) == slot + 1:
                change_slot = False
            if change_slot is True:
                logger.info(f"Activating slot {slot}, current is {current_state.get('active_slot', 'UNKNOWN')}")
                config_cmds.insert(0, MinidspBeqCommandGenerator.activate(slot))
        formatted = '\n'.join(config_cmds)
        logger.info(f"\n{formatted}")
        with tmp_file(config_cmds) as file_name:
            kwargs = {'retcode': None} if self.__ignore_retcode else {}
            logger.info(f"Sending {len(config_cmds)} commands to slot {slot} via {file_name} {kwargs if kwargs else ''}")
            start = time.time()
            code, stdout, stderr = self.__runner['-f', file_name].run(timeout=self.__cmd_timeout, **kwargs)
            end = time.time()
            logger.info(f"Sent {len(config_cmds)} commands to slot {slot} in {to_millis(start, end)}ms - result is {code}")


class MinidspSlotState(SlotState):

    def __init__(self, slot_id: str):
        super().__init__(slot_id)
        self.gain1 = 0.0
        self.mute1 = False
        self.gain2 = 0.0
        self.mute2 = False

    def clear(self):
        super().clear()
        self.gain1 = 0.0
        self.gain2 = 0.0
        self.mute1 = False
        self.mute2 = False

    def set_gain(self, channel: Optional[int], value: float):
        if channel is None:
            self.gain1 = value
            self.gain2 = value
        else:
            if channel == 1:
                self.gain1 = value
            elif channel == 2:
                self.gain2 = value
            else:
                raise ValueError(f'Unknown channel {channel} for slot {self.slot_id}')

    def mute(self, channel: Optional[int]):
        self.__do_mute(channel, True)

    def __do_mute(self, channel: Optional[int], value: bool):
        if channel is None:
            self.mute1 = value
            self.mute2 = value
        else:
            if channel == 1:
                self.mute1 = value
            elif channel == 2:
                self.mute2 = value
            else:
                raise ValueError(f'Unknown channel {channel} for slot {self.slot_id}')

    def unmute(self, channel: Optional[int]):
        self.__do_mute(channel, False)

    def merge_with(self, state: dict) -> bool:
        super().merge_with(state)
        if 'gain1' in state:
            self.gain1 = float(state['gain1'])
        if 'gain2' in state:
            self.gain2 = float(state['gain2'])
        if 'mute1' in state:
            self.mute1 = bool(state['mute1'])
        if 'mute2' in state:
            self.mute2 = bool(state['mute2'])
        return True

    def as_dict(self) -> dict:
        sup = super().as_dict()
        return {
            **sup,
            'gain1': self.gain1,
            'gain2': self.gain2,
            'mute1': self.mute1,
            'mute2': self.mute2,
            'canActivate': True
        }

    def __repr__(self):
        return f"{super().__repr__()} - 1: {self.gain1:.2f}/{self.mute1} 2: {self.gain1:.2f}/{self.mute2}"


class MinidspBeqCommandGenerator:

    @staticmethod
    def activate(slot: int) -> str:
        return f"config {slot}"

    @staticmethod
    def filt(entry: Optional[Catalogue]):
        # input <channel> peq <index> set -- <b0> <b1> <b2> <a1> <a2>
        # input <channel> peq <index> bypass [on|off]
        cmds = []
        for c in range(2):
            idx = 0
            if entry:
                for f in entry.filters:
                    bq: dict = f['biquads']['96000']
                    coeffs: List[str] = bq['b'] + bq['a']
                    if len(coeffs) != 5:
                        raise ValueError(f"Invalid coeff count {len(coeffs)} at idx {idx}")
                    else:
                        cmds.append(MinidspBeqCommandGenerator.bq(c, idx, coeffs))
                        cmds.append(MinidspBeqCommandGenerator.bypass(c, idx, False))
                        idx += 1
            for i in range(idx, 10):
                cmds.append(MinidspBeqCommandGenerator.bypass(c, i, True))
        return cmds

    @staticmethod
    def bq(channel: int, idx: int, coeffs):
        return f"input {channel} peq {idx} set -- {' '.join(coeffs)}"

    @staticmethod
    def bypass(channel: int, idx: int, bypass: bool):
        return f"input {channel} peq {idx} bypass {'on' if bypass else 'off'}"

    @staticmethod
    def mute(state: bool, slot: Optional[int], channel: Optional[int]):
        '''
        Generates commands to mute the configuration.
        :param state: mute if true otherwise unmute.
        :param slot: the target slot, if not set apply to the master control.
        :param channel: the channel, applicable only if slot is set, if not set apply to both input channels.
        :return: the commands.
        '''
        state_cmd = 'on' if state else 'off'
        if slot is not None:
            cmds = []
            if channel is None:
                cmds.append(f"input 0 mute {state_cmd}")
                cmds.append(f"input 1 mute {state_cmd}")
            else:
                cmds.append(f"input {channel} mute {state_cmd}")
            return cmds
        else:
            return [f"mute {state_cmd}"]

    @staticmethod
    def gain(gain: float, slot: Optional[int], channel: Optional[int]):
        '''
        Generates commands to set gain.
        :param gain: the gain to set.
        :param slot: the target slot, if not set apply to the master control.
        :param channel: the channel, applicable only if slot is set, if not set apply to both input channels.
        :return: the commands.
        '''
        if slot is not None:
            if not -72.0 <= gain <= 12.0:
                raise InvalidRequestError(f"Input gain {gain:.2f} out of range (>= -72.0 and <= 12.0)")
            cmds = []
            if channel is None:
                cmds.append(f"input 0 gain -- {gain:.2f}")
                cmds.append(f"input 1 gain -- {gain:.2f}")
            else:
                cmds.append(f"input {channel} gain -- {gain:.2f}")
            return cmds
        else:
            if not -127.0 <= gain <= 0.0:
                raise InvalidRequestError(f"Master gain {gain:.2f} out of range (>= -127.0 and <= 0.0)")
            return [f"gain -- {gain:.2f}"]


@contextmanager
def tmp_file(cmds: List[str]):
    import tempfile
    tmp_name = None
    try:
        f = tempfile.NamedTemporaryFile(mode='w+', delete=False)
        for cmd in cmds:
            f.write(cmd)
            f.write('\n')
        tmp_name = f.name
        f.close()
        yield tmp_name
    finally:
        if tmp_name:
            os.unlink(tmp_name)


def to_millis(start, end, precision=1):
    '''
    Calculates the differences in time in millis.
    :param start: start time in seconds.
    :param end: end time in seconds.
    :return: delta in millis.
    '''
    return round((end - start) * 1000, precision)



import logging
from typing import List, Tuple, Optional

import math
from flask import request
from flask_restx import Resource, Namespace, fields

from ezbeq.catalogue import CatalogueProvider, Catalogue
from ezbeq.device import DeviceStateHolder, DeviceBridge, InvalidRequestError, DeviceState

logger = logging.getLogger('ezbeq.devices')


def delete_filter(bridge: DeviceBridge, state: DeviceStateHolder, slot: str):
    '''
    Clears the slot.
    :param bridge: the bridge to the device.
    :param state: state of the bridge.
    :param slot: the slot.
    :return: current state after clearing, 200 if cleared or 500 if unable to load
    '''
    logger.info(f"Clearing Slot {slot}")
    try:
        bridge.clear_filter(slot)
        state.clear(slot)
        return state.get().as_dict(), 200
    except Exception as e:
        logger.exception(f"Failed to clear Slot {slot}")
        state.error(slot)
        return state.get().as_dict(), 500


def load_filter(catalogue: List[Catalogue], bridge: DeviceBridge, state: DeviceStateHolder, entry_id: str,
                slot: str) -> Tuple[dict, int]:
    '''
    Attempts to find the supplied entry in the catalogue and load it into the given slot.
    :param catalogue: the catalogue.
    :param bridge: the bridge to the device.
    :param state: state of the bridge.
    :param entry_id: the catalogue entry id.
    :param slot: the slot.
    :return: current state after load, 200 if loaded, 400 if no such entry, 500 if unable to load
    '''
    logger.info(f"Sending {entry_id} to Slot {slot}")
    match: Catalogue = next((c for c in catalogue if c.idx == entry_id), None)
    if not match:
        logger.warning(f"No title with ID {entry_id} in catalogue")
        return {'message': 'Title not found, please refresh.'}, 404
    try:
        bridge.load_filter(slot, match)
        state.set_loaded_entry(slot, match)
        return state.get().as_dict(), 200
    except InvalidRequestError as e:
        logger.exception(f"Invalid request {entry_id} to Slot {slot}")
        return state.get().as_dict(), 400
    except Exception as e:
        logger.exception(f"Failed to write {entry_id} to Slot {slot}")
        state.error(slot)
        return state.get().as_dict(), 500


def activate_slot(bridge: DeviceBridge, state: DeviceStateHolder, slot: str):
    '''
    Activates the slot.
    :param bridge: the bridge to the device.
    :param state: state of the bridge.
    :param slot: the slot.
    :return: current state after activation, 200 if activated or 500 if unable to activate
    '''
    logger.info(f"Activating Slot {slot}")
    try:
        bridge.activate(slot)
        state.activate(slot)
        return state.get().as_dict(), 200
    except InvalidRequestError as e:
        logger.exception(f"Invalid slot {slot}")
        return state.get().as_dict(), 400
    except Exception as e:
        logger.exception(f"Failed to activate Slot {slot}")
        state.error(slot)
        return state.get().as_dict(), 500


def mute_device(bridge: DeviceBridge, state: DeviceStateHolder, device_name: str, slot: Optional[str], value: bool,
                channel: Optional[int] = None):
    '''
    Mutes or unmutes a particular aspect of the device.
    :param bridge: the bridge to the device.
    :param state: state of the bridge.
    :param device_name: device to affect.
    :param slot: optional slot id.
    :param value: whether to mute (true) or unmute (false)
    :param channel: optional input channel id.
    :return: current state after making changes, 200 if updated or 500 if unable to update
    '''
    try:
        if value:
            bridge.mute(slot, channel)
            if slot is None:
                state.master_mute = value
            else:
                state.mute_slot(slot, channel)
        else:
            bridge.unmute(slot, channel)
            if slot is None:
                state.master_mute = value
            else:
                state.unmute_slot(slot, channel)
        return state.get().as_dict(), 200
    except InvalidRequestError as e:
        logger.exception(f"Invalid mute request {slot} {channel} {value}")
        return state.get().as_dict(), 400
    except Exception as e:
        logger.exception(f"Failed mute channel {slot}")
        return state.get().as_dict(), 500


def set_gain(bridge: DeviceBridge, state: DeviceStateHolder, device_name: str, slot: Optional[str], value: float,
             channel: Optional[int] = None):
    '''
    Sets gain on a particular aspect of the device.
    :param bridge: the bridge to the device.
    :param state: state of the bridge.
    :param device_name: device to affect.
    :param slot: optional slot id.
    :param value: the gain level to set.
    :param channel: optional input channel id.
    :return: current state after making changes, 200 if updated or 500 if unable to update
    '''
    try:
        bridge.set_gain(slot, channel, value)
        if slot is None:
            state.master_volume = value
        else:
            state.set_slot_gain(slot, channel, value)
        return state.get().as_dict(), 200
    except InvalidRequestError as e:
        logger.exception(f"Unable to set gain for {slot} {channel} {value}")
        return state.get().as_dict(), 400
    except Exception as e:
        logger.exception(f"Failed mute channel {slot}")
        return state.get().as_dict(), 500


api = Namespace('devices', description='Device related operations')


@api.route('')
class Devices(Resource):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__state: DeviceStateHolder = kwargs['device_state']
        self.__bridge: DeviceBridge = kwargs['device_bridge']

    def get(self):
        self.__state.initialise(self.__bridge)
        return self.__state.get().as_dict()


slot_model = api.model('Slot', {
    'id': fields.String(required=True),
    'active': fields.Boolean(required=False),
    'gain1': fields.Float(required=False),
    'gain2': fields.Float(required=False),
    'mute1': fields.Boolean(required=False),
    'mute2': fields.Boolean(required=False),
    'entry': fields.String(required=False)
})

device_model = api.model('Device', {
    'mute': fields.Boolean(required=False),
    'masterVolume': fields.Float(required=False),
    'slots': fields.List(fields.Nested(slot_model), required=False)
})


@api.route('/<string:device_name>', defaults={'device_name': 'master'})
class Device(Resource):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__state: DeviceStateHolder = kwargs['device_state']
        self.__bridge: DeviceBridge = kwargs['device_bridge']
        self.__catalogue_provider: CatalogueProvider = kwargs['catalogue']

    @api.expect(device_model, validate=True)
    def patch(self, device_name: str):
        current_state: DeviceState = self.__state.get()
        if current_state:
            payload = request.get_json()
            logger.info(f"Processing PATCH for {device_name} {payload}")
            if 'slots' in payload:
                for slot in payload['slots']:
                    state, cmd_code = self.__update_slot(slot, current_state, device_name)
                    if cmd_code != 200:
                        return state, cmd_code
            if 'mute' in payload and payload['mute'] != current_state.mute:
                state, cmd_code = mute_device(self.__bridge, self.__state, device_name, None, payload['mute'], None)
                if cmd_code != 200:
                    return state, cmd_code
            if 'masterVolume' in payload and not math.isclose(payload['masterVolume'], current_state.master_volume):
                state, cmd_code = set_gain(self.__bridge, self.__state, device_name, None, payload['masterVolume'],
                                           None)
                if cmd_code != 200:
                    return state, cmd_code
            return self.__state.get().as_dict(), 200
        else:
            return f"Unknown device {device_name}", 404

    def __update_slot(self, slot: dict, current_state: DeviceState, device_name: str) -> Tuple[dict, int]:
        state = {}
        cmd_code = 400
        try:
            current_state.get_slot(slot['id'])
        except StopIteration as e:
            return {'msg': f"Unknown slot {slot['id']} in device {device_name}"}, 400
        if 'gain1' in slot:
            state, cmd_code = set_gain(self.__bridge, self.__state, device_name, slot['id'], slot['gain1'], 1)
            if cmd_code != 200:
                return state, cmd_code
        if 'gain2' in slot:
            state, cmd_code = set_gain(self.__bridge, self.__state, device_name, slot['id'], slot['gain2'], 2)
            if cmd_code != 200:
                return state, cmd_code
        if 'mute1' in slot:
            state, cmd_code = mute_device(self.__bridge, self.__state, device_name, slot['id'], slot['mute1'], 1)
            if cmd_code != 200:
                return state, cmd_code
        if 'mute2' in slot:
            state, cmd_code = mute_device(self.__bridge, self.__state, device_name, slot['id'], slot['mute1'], 2)
            if cmd_code != 200:
                return state, cmd_code
        if 'entry' in slot and slot['entry']:
            state, cmd_code = load_filter(self.__catalogue_provider.catalogue, self.__bridge, self.__state,
                                          slot['entry'], slot['id'])
            if cmd_code != 200:
                return state, cmd_code
        if 'active' in slot:
            state, cmd_code = activate_slot(self.__bridge, self.__state, slot['id'])
            if cmd_code != 200:
                return state, cmd_code
        return state, cmd_code


@api.route('/<string:device_name>/config/<string:slot>/active')
@api.doc(params={
    'device_name': 'The dsp device name',
    'slot': 'The dsp configuration to activate, available values depend on the DSP device (1-4 for MiniDSP 2x4HD)'
})
class ActiveSlot(Resource):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__bridge: DeviceBridge = kwargs['device_bridge']
        self.__state: DeviceStateHolder = kwargs['device_state']

    def put(self, device_name: str, slot: str) -> Tuple[dict, int]:
        return activate_slot(self.__bridge, self.__state, slot)


gain_model = api.model('Gain', {
    'gain': fields.Float
})


@api.route('/<string:device_name>/gain/<string:slot>/<int:channel>')
@api.route('/<string:device_name>/gain/<string:slot>')
@api.route('/<string:device_name>/gain')
@api.doc(params={
    'device_name': 'The dsp device name',
    'slot': 'The dsp configuration to set the gain on, available values depend on the DSP device (1-4 for MiniDSP'
            '2x4HD). If unset, the master gain is changed.',
    'channel': 'sets the gain on the specified input channel or all inputs if not set'
})
class SetGain(Resource):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__bridge: DeviceBridge = kwargs['device_bridge']
        self.__state: DeviceStateHolder = kwargs['device_state']

    @api.expect(gain_model, validate=True)
    def put(self, device_name: str, slot: Optional[str] = None, channel: Optional[int] = None) -> Tuple[dict, int]:
        return set_gain(self.__bridge, self.__state, device_name, slot, request.get_json()['gain'], channel)

    def delete(self, device_name: str, slot: Optional[str] = None, channel: Optional[int] = None) -> Tuple[dict, int]:
        return set_gain(self.__bridge, self.__state, device_name, slot, 0.0, channel)


@api.route('/<string:device_name>/mute/<string:slot>/<int:channel>')
@api.route('/<string:device_name>/mute/<string:slot>')
@api.route('/<string:device_name>/mute')
@api.doc(params={
    'device_name': 'The dsp device name',
    'slot': 'The dsp configuration to mute, available values depend on the DSP device (1-4 for MiniDSP 2x4HD). '
            'If unset, the entire device is muted.',
    'channel': 'mutes or unmutes the specified input channel or all inputs if not set'
})
class Mute(Resource):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__bridge: DeviceBridge = kwargs['device_bridge']
        self.__state: DeviceStateHolder = kwargs['device_state']

    def put(self, device_name: str, slot: Optional[str] = None, channel: Optional[int] = None) -> Tuple[dict, int]:
        return mute_device(self.__bridge, self.__state, device_name, slot, True, channel)

    def delete(self, device_name: str, slot: Optional[str] = None, channel: Optional[int] = None) -> Tuple[dict, int]:
        return mute_device(self.__bridge, self.__state, device_name, slot, False, channel)


filter_model = api.model('Filter', {
    'entryId': fields.String
})


@api.route('/<string:device_name>/filter/<string:slot>')
@api.doc(params={
    'device_name': 'The dsp device name',
    'slot': 'The dsp configuration to load into, available values depend on the DSP device (1-4 for MiniDSP 2x4HD)'
})
class ManageFilter(Resource):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__bridge: DeviceBridge = kwargs['device_bridge']
        self.__catalogue_provider: CatalogueProvider = kwargs['catalogue']
        self.__state: DeviceStateHolder = kwargs['device_state']

    @api.expect(filter_model, validate=True)
    def put(self, device_name: str, slot: str) -> Tuple[dict, int]:
        return load_filter(self.__catalogue_provider.catalogue, self.__bridge, self.__state,
                           request.get_json()['entryId'], slot)

    def delete(self, device_name: str, slot: str) -> Tuple[dict, int]:
        return delete_filter(self.__bridge, self.__state, slot)


# legacy API, deprecated
device_api = Namespace('device')


@device_api.route('/<string:slot>')
@device_api.deprecated
class DeviceSender(Resource):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__bridge: DeviceBridge = kwargs['device_bridge']
        self.__catalogue_provider: CatalogueProvider = kwargs['catalogue']
        self.__state: DeviceStateHolder = kwargs['device_state']

    def put(self, slot: str):
        '''
        Handles update commands for the device. The following rules apply
        * slot is a string in the 1-4 range representing minidsp config slots except for mute/gain commands which translate 0 to the master controls
        * command is one of load, activate, mute, gain
        * the payload may be a single dict or a list of dicts (for multiple commands)
        * mute: requires channel and value where channel is 0 (meaning 1 and 2) or master, 1 or 2 and refers to the input channel, value is on or off
        * gain: requires channel and value where channel is 0 (meaning 1 and 2) or master, 1 or 2 and refers to the input channel, value is the dB level to set
        * activate: requires no further data
        * load: requires id which refers to catalogue.id
        * mute and gain also accept channel = master which is used to override the provided slot and target the command at the master controls
        :param slot: the slot to target.
        :return: the device state, http status code.
        '''
        payload = request.get_json()
        if isinstance(payload, list):
            result = None
            for p in payload:
                result, _ = self.__handle_command(slot, p)
            return self.__state.get().as_dict(), 200 if result else 500
        elif isinstance(payload, dict):
            return self.__handle_command(slot, payload)
        return self.__state.get().as_dict(), 404

    def __handle_command(self, slot: str, payload: dict):
        if 'command' in payload:
            cmd = payload['command']
            if cmd == 'load':
                if 'id' in payload:
                    return load_filter(self.__catalogue_provider.catalogue, self.__bridge, self.__state, payload['id'],
                                       slot)
            elif cmd == 'activate':
                return activate_slot(self.__bridge, self.__state, slot)
            elif cmd == 'mute' or cmd == 'gain':
                if 'value' in payload:
                    channel = payload['channel'] if 'channel' in payload else None
                    if channel == 'master':
                        slot = None
                        channel = None
                    elif channel == '0':
                        channel = None
                    channel = int(channel) if channel is not None else None
                    if cmd == 'mute':
                        return mute_device(self.__bridge, self.__state, 'NOP', slot,
                                           False if payload['value'] == 'off' else True, channel)
                    else:
                        return set_gain(self.__bridge, self.__state, 'NOP', slot, round(float(payload['value']), 2),
                                        channel)
        return None, 400

    def delete(self, slot):
        return delete_filter(self.__bridge, self.__state, slot)
